from argparse import ArgumentParser

import pandas as pd
from bilby.core.prior import PriorDict
from bilby.core.utils import logger

from .utils import prior_conversion


def create_parser():
    parser = ArgumentParser(prog="GWPopulation pipe create injections")
    parser.add_argument("--prior-file", type=str, help="The prior file to sample from.")
    parser.add_argument("--n-simulation", type=int, help="The number of injections")
    parser.add_argument("--injection-file", type=str, help="The output file.")
    return parser


def create_injections(prior_file, n_simulation):
    priors = PriorDict(prior_file)
    priors.conversion_function = prior_conversion
    logger.info(f"Drawing {n_simulation} samples from {prior_file}.")
    samples = pd.DataFrame(priors.sample(n_simulation))
    return samples


def main():
    parser = create_parser()
    args = parser.parse_args()
    samples = create_injections(
        prior_file=args.prior_file, n_simulation=args.n_simulation
    )
    samples.to_json(args.injection_file)
    logger.info(f"Saving samples to {args.injection_file}.")



from gwpopulation.utils import truncnorm


def all_the_gaussians(
    dataset,
    mean_mass,
    sigma_mass,
    mean_magnitude,
    sigma_magnitude,
    mean_tilt,
    sigma_tilt,
    mean_redshift,
    sigma_redshift,
):
    """
    A simple model to use a test case for numpyro.

    We use a truncated normal for all variables and apply a log-transform to the mass
    """
    from gwpopulation.utils import xp

    prob = truncnorm(dataset["redshift"], mean_redshift, sigma_redshift, low=0, high=3)
    for ii in [1, 2]:
        mass = dataset[f"mass_{ii}"]
        log_mass = xp.log10(mass)
        prob *= truncnorm(log_mass, mean_mass, sigma_mass, low=0, high=2.5) / mass
        prob *= truncnorm(
            dataset[f"a_{ii}"], mean_magnitude, sigma_magnitude, low=0, high=1
        )
        prob *= truncnorm(
            dataset[f"cos_tilt_{ii}"], mean_tilt, sigma_tilt, low=-1, high=1
        )
    return prob



"""
Functions for collecting input samples from a range of sources and computing
the fiducial prior for the appropriate parameters.

The module provides the `gwpopulation_pipe_collection` executable.

In order to use many of the other functions you will need a class that provides
various attributes specified in the `gwpopulation_pipe` parser.
"""

#!/usr/bin/env python3

import json
import os
import re

import glob
import h5py
import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from bilby.core.utils import logger
from bilby_pipe.utils import convert_string_to_dict
from gwpopulation.backend import set_backend

from .data_simulation import simulate_posteriors
from .analytic_spin_prior import prior_chieff_chip_isotropic
from .parser import create_parser
from .utils import get_cosmology
from .vt_helper import dump_injection_data

matplotlib.rcParams["text.usetex"] = False


def euclidean_redshift_prior(redshift, cosmology="Planck15_LAL"):
    r"""
    Evaluate the redshift prior assuming a Euclidean universe.

    See Appendix C of `Abbott et al. <https://arxiv.org/pdf/1811.12940.pdf>`_.

    .. math::

        p(z) \propto d^2_L \frac{dd_L}{dz}

    Parameters
    ----------
    redshift: array_like
        The redshift values to evaluate the prior for.
    cosmology: str
        The name of the cosmology, default is `Planck15_LAL`

    """
    cosmo = get_cosmology(cosmology)

    luminosity_distance = cosmo.luminosity_distance(redshift)
    return luminosity_distance**2 * cosmo.dDLdz(redshift)


def euclidean_distance_prior(redshift):
    logger.warning(
        "The euclidean_distance_prior function is deprecated, "
        "use euclidean_redshift_prior instead."
    )
    return euclidean_redshift_prior(redshift)


def cosmological_redshift_prior(redshift, cosmology="Planck15_LAL"):
    r"""
    Evaluate the redshift prior assuming a cosmological universe.

    .. math::

        p(z) \propto \frac{4\pi}{1 + z}\frac{dV_C}{dz}

    Parameters
    ----------
    redshift: array_like
        The redshift values to evaluate the prior for.
    cosmology: str
        The name of the cosmology, default is `Planck15_LAL`

    """
    cosmo = get_cosmology(cosmology)

    return cosmo.differential_comoving_volume(redshift) * 4 * np.pi / (1 + redshift)


def distance_prior(redshift_prior, luminosity_distance, cosmology="Planck15_LAL"):
    r"""
    Calculate the prior on luminosity distance given a redshift prior assuming the given
    cosmology.

    .. math::

        p(d_L) = p(z)\frac{dz}{dd_L}

    Parameters
    ----------
    redshift_prior: callable
        The redshift prior function.
    luminosity_distance: array_like
        The luminosity distance values to evaluate the prior for.
    cosmology: str
        The name of the cosmology, default is `Planck15_LAL`

    """
    from wcosmo import available, z_at_value
    from wcosmo.utils import disable_units

    disable_units()
    cosmo = get_cosmology(cosmology)

    redshift = z_at_value(cosmo.luminosity_distance, luminosity_distance)
    return redshift_prior(redshift) / cosmo.dDLdz(redshift)


def aligned_spin_prior(spin):
    r"""
    The standard prior for aligned spin assuming the spin prior extends to maximal.

    .. math::

        p(\chi) = \frac{1}{2} \log(|\chi|)

    Parameters
    ----------
    spin: array_like
        The aligned spin values to evaluate the prior for.

    Returns
    -------
    prior: array_like
        The prior evaluated at the input spin.
    """
    return -np.log(np.abs(spin)) / 2


def primary_mass_to_chirp_mass_jacobian(samples):
    r"""
    Compute the Jacobian for the primary mass to chirp mass transformation.

    .. math::

        \frac{d m_c}{d m_1} = \frac{q^{3/5}}{(1 + q)^{1/5}}

    Parameters
    ----------
    samples: dict
        Samples containing `mass_1` and `mass_ratio`.

    Returns
    -------
    jacobian: array_like
        The Jacobian for the transformation.
    """
    return (1 + samples["mass_ratio"]) ** 0.2 / samples["mass_ratio"] ** 0.6


def replace_keys(posts):
    """
    Map the keys from legacy names to the `GWPopulation` standards.

    Parameters
    ----------
    posts: dict
        Dictionary of `pd.DataFrame` objects

    Returns
    -------
    new_posts: dict
        Updated posteriors.

    """
    _mapping = dict(
        mass_1="m1_source",
        mass_2="m2_source",
        mass_ratio="q",
        a_1="a1",
        a_2="a2",
        cos_tilt_1="costilt1",
        cos_tilt_2="costilt2",
        redshift="redshift",
        chi_eff="chi_eff",
        chi_p="chi_p",
    )
    new_posts = dict()
    for name in posts:
        post = posts[name]
        new = pd.DataFrame()
        for key in _mapping:
            if _mapping[key] in post:
                new[key] = post[_mapping[key]]
            elif key in post:
                new[key] = post[key]
            else:
                new[key] = 0
        new_posts[name] = new
    return new_posts


def evaluate_prior(posts, args, dataset, meta):
    """
    Evaluate the prior distribution for the input posteriors.

    Parameters
    ----------
    posts: dict
        Dictionary of `pd.DataFrame` objects containing the posteriors.
    args:
        Input args containing the prior specification.
    dataset: str
        The dataset label to evaluate the prior for (i.e. "O4a"). Should be a key in `args.sample_regex`.
    meta: dict
        The per-event metadata, including e.g., the cosmology used for the analysis.

    Returns
    -------
    posts: dict
        The input dictionary, modified in place.
    """
    if "redshift" in args.parameters or "luminosity_distance" in args.parameters:
        if args.distance_prior[dataset].lower() == "comoving":
            logger.info(
                "Using uniform in the comoving source frame distance prior for all events."
            )
            redshift_prior = cosmological_redshift_prior
        elif args.distance_prior[dataset].lower() == "euclidean":
            logger.info("Using Euclidean distance prior for all events.")
            redshift_prior = euclidean_redshift_prior
        elif args.distance_prior[dataset].lower() == "none":
            redshift_prior = lambda x, *args, **kwargs: x**0
        else:
            raise ValueError(f"Redshift prior {args.distance_prior} not recognized")
    else:
        logger.warning(
            f"No redshift and no luminosity distance present for {name}, cannot evaluate distance prior weight"
        )

    if args.mass_prior[dataset].lower() == "flat-detector-components":
        logger.info("Assuming flat in detector frame mass prior for all events.")
        if "mass_1_detector" in args.parameters:
            logger.debug(
                f"no (1+z) factor since the priors are now in detector coordinate"
            )
    elif args.mass_prior[dataset].lower() == "flat-detector-chirp-mass-ratio":
        logger.info("Assuming chirp mass prior for all events.")
    elif args.mass_prior[dataset].lower() not in ["flat-source-components", "none"]:
        raise ValueError(f"Mass prior {args.mass_prior[dataset]} not recognized.")

    if args.spin_prior[dataset].lower() == "component":
        logger.info("Assuming uniform in component spin prior for all events.")
    elif args.spin_prior[dataset].lower() != "none":
        raise ValueError(f"Spin prior {args.spin_prior[dataset]} not recognized.")

    if "mass_ratio" in args.parameters:
        logger.info(
            "Model is defined in terms of mass ratio, adjusting prior accordingly."
        )
    if "chirp_mass" in args.parameters or "chip_mass_detector" in args.parameters:
        logger.info(
            "Model is defined in terms of chirp mass, adjusting prior accordingly."
        )

    for name in posts:
        post = posts[name]
        cosmology = meta.get(name, dict()).get("cosmology", None)
        if cosmology is None:
            cosmology = "Planck15_LAL"
        logger.info(f"Using {cosmology} cosmology for {name}")
        post["prior"] = 1
        if "redshift" in args.parameters:
            post["prior"] *= redshift_prior(post["redshift"], cosmology=cosmology)
        elif "luminosity_distance" in args.parameters:
            post["prior"] *= distance_prior(
                redshift_prior, post["luminosity_distance"], cosmology=cosmology
            )

        if "mass_1" in args.parameters:
            if args.mass_prior[dataset].lower() == "flat-detector-components":
                post["prior"] *= (1 + post["redshift"]) ** 2
            elif args.mass_prior[dataset].lower() == "flat-detector-chirp-mass-ratio":
                post["prior"] /= (
                    post["mass_1"]
                    / (1 + post["redshift"])
                    * primary_mass_to_chirp_mass_jacobian(post)
                )
            if "mass_ratio" in args.parameters:
                post["prior"] *= post["mass_1"]
        elif "mass_1_detector" in args.parameters:
            if args.mass_prior[dataset].lower() == "flat-detector-chirp-mass-ratio":
                post["prior"] /= post[
                    "mass_1_detector"
                ] * primary_mass_to_chirp_mass_jacobian(post)
            if "mass_ratio" in args.parameters:
                post["prior"] *= post["mass_1_detector"]

        if "chirp_mass" in args.parameters or "chirp_mass_detector" in args.parameters:
            post["prior"] *= primary_mass_to_chirp_mass_jacobian(post)

        if args.spin_prior[dataset].lower() == "component":
            post["prior"] /= 4

        if "chi_eff" in args.parameters:
            if "chi_p" in args.parameters:
                post["prior"] *= prior_chieff_chip_isotropic(
                    post["chi_eff"], post["chi_p"], post["mass_ratio"]
                )
            else:
                logger.warning("chi_eff prior specified without chi_p.")

        if "chi_1" in args.parameters:
            post["prior"] *= aligned_spin_prior(post["chi_1"])
        if "chi_2" in args.parameters:
            post["prior"] *= aligned_spin_prior(post["chi_2"])

    return posts


def load_posterior_from_meta_file(filename, labels=None):
    """
    Load a posterior from a `PESummary` meta file. The poseterior samples are expected to follow Bilby naming conventions.

    Parameters
    ----------
    filename: str
    labels: list
        The labels to search for in the file in order of precedence.

    Returns
    -------
    posterior: pd.DataFrame
    meta_data: dict
        Dictionary containing the run label that was loaded.

    """
    _mapping = dict(
        chirp_mass="chirp_mass_source",
        mass_1="mass_1_source",
        mass_2="mass_2_source",
        chirp_mass_detector="chirp_mass",
        mass_1_detector="mass_1",
        mass_2_detector="mass_2",
        mass_ratio="mass_ratio",
        redshift="redshift",
        luminosity_distance="luminosity_distance",
        a_1="a_1",
        a_2="a_2",
        cos_tilt_1="cos_tilt_1",
        cos_tilt_2="cos_tilt_2",
        chi_1="spin_1z",
        chi_2="spin_2z",
        chi_eff="chi_eff",
        chi_p="chi_p",
    )
    load_map = dict(
        json=load_meta_file_from_json,
        h5=load_meta_file_from_hdf5,
        hdf5=load_meta_file_from_hdf5,
        dat=load_samples_from_csv,
    )
    if labels is None:
        labels = ["PrecessingSpinIMRHM", "PrecessingSpin"]
    if not os.path.exists(filename):
        raise FileNotFoundError(f"{filename} does not exist")
    extension = os.path.splitext(filename)[1][1:]
    _posterior, label, cosmology = load_map[extension](filename=filename, labels=labels)
    keys = [key for key, value in _mapping.items() if value in _posterior]
    posterior = pd.DataFrame({key: _posterior[_mapping[key]] for key in keys})
    _attempt_to_fill_posterior(posterior, cosmology)
    meta_data = dict(label=label, cosmology=cosmology)
    logger.info(f"Loaded {label} from {filename}.")
    return posterior, meta_data


def _attempt_to_fill_posterior(posterior, cosmology=None):
    """
    Attempt to add missing variables to the posterior.
    This is mostly for CI testing where the data doesn't contain all the variables.
    """
    from wcosmo import available, z_at_value
    from wcosmo.utils import disable_units

    disable_units()
    if cosmology is None:
        cosmology = "Planck15_LAL"
    logger.info(f"Assuming {cosmology} cosmology")
    cosmo = get_cosmology(cosmology)

    if "redshift" in posterior and "luminosity_distance" not in posterior:
        posterior["luminosity_distance"] = cosmo.luminosity_distance(
            posterior["redshift"]
        )
    elif "luminosity_distance" in posterior and "redshift" not in posterior:
        posterior["redshift"] = z_at_value(
            cosmo.redshift, posterior["luminosity_distance"]
        )
    elif "redshift" not in posterior:
        return

    for var in ["mass_1", "mass_2", "chirp_mass"]:
        if var in posterior and f"{var}_detector" not in posterior:
            posterior[f"{var}_detector"] = posterior[var] * (1 + posterior["redshift"])
        elif f"{var}_detector" in posterior and var not in posterior:
            posterior[var] = posterior[f"{var}_detector"] / (1 + posterior["redshift"])


def load_meta_file_from_hdf5(filename, labels):
    """
    Load the posterior from a `hdf5` `PESummary` file.
    See `load_posterior_from_meta_file`.
    """
    new_style = True
    with h5py.File(filename, "r") as data:
        if "posterior_samples" in data.keys():
            new_style = False
            data = data["posterior_samples"]
        label = None
        for _label in labels:
            if _label in data.keys():
                label = _label
                break
        if label is None:
            for key in data.keys():
                if "posterior_samples" in data[key]:
                    label = key
                    break
        if label is None:
            raise ValueError(f"Could not find posterior samples in {filename}")
        try:
            cosmology = data[label]["meta_data"]["meta_data"]["cosmology"][0].decode()
        except (AttributeError, IndexError, KeyError):
            cosmology = None
        if new_style:
            posterior = pd.DataFrame(data[label]["posterior_samples"][:])
        elif hasattr(data[label], "keys"):
            posterior = pd.DataFrame(
                data[label]["samples"][:],
                columns=[key.decode() for key in data[label]["parameter_names"][:]],
            )
        else:
            posterior = pd.DataFrame(data[label][:])
        return posterior, label, cosmology


def load_meta_file_from_json(filename, labels):
    """
    Load the posterior from a `json` `PESummary` file.
    See `load_posterior_from_meta_file`.
    """
    with open(filename, "r") as ff:
        data = json.load(ff)
    samples = data["posterior_samples"]
    del data
    label = list(samples.keys())[0]
    for _label in labels:
        if _label in samples:
            label = _label
            break
    try:
        cosmology = data[label]["meta_data"]["meta_data"]["cosmology"]
    except (AttributeError, IndexError, KeyError):
        cosmology = None
    posterior = pd.DataFrame(
        samples[label]["samples"], columns=samples[label]["parameter_names"]
    )
    return posterior, label, cosmology


def load_samples_from_csv(filename, *args, **kwargs):
    """
    Load posterior samples from a csd file.
    This is just a wrapper to `pd.read_csv` assuming tab separation.

    Parameters
    ----------
    filename: str
    args: unused
    kwargs: unused

    Returns
    -------
    posterior: `pd.DataFrame`
    meta_data: None
    cosmology: None
    """
    posterior = pd.read_csv(filename, sep="\t")
    return posterior, None, None


def _load_batch_of_meta_files(regex, label, labels=None, keys=None, ignore=None):
    if ignore is None:
        ignore = list()
    if keys is None:
        keys = [
            "mass_1",
            "mass_ratio",
            "a_1",
            "a_2",
            "cos_tilt_1",
            "cos_tilt_2",
            "chi_1",
            "chi_2",
            "redshift",
            "chi_eff",
            "chi_p",
        ]
    posteriors = dict()
    meta_data = dict()
    all_files = glob.glob(regex)
    all_files.sort()
    logger.info(f"Found {len(all_files)} {label} events in standard format.")
    for posterior_file in all_files:
        drop = False
        for label in ignore:
            if label in posterior_file:
                drop = True
                break
        if drop:
            logger.info(f"Ignoring {posterior_file}.")
            continue
        try:
            new_posterior, data = load_posterior_from_meta_file(
                posterior_file, labels=labels
            )
        except (TypeError, ValueError) as e:
            logger.info(f"Failed to load {posterior_file} with {type(e)}: {e}.")
            continue
        if all([key in new_posterior for key in keys]):
            meta_data[posterior_file] = data
            new_posterior = new_posterior[keys]
            if "mass_ratio" in new_posterior:
                new_posterior["mass_ratio"] = np.minimum(
                    new_posterior["mass_ratio"],
                    1 / new_posterior["mass_ratio"],
                )
            posteriors[posterior_file] = new_posterior
        else:
            logger.info(f"Posterior has keys {new_posterior.keys()}.")
    return posteriors, meta_data


def load_all_events(args, save_meta_data=True, ignore=None):
    """
    Load posteriors for some/all events.

    Parameters
    ----------
    args: argparse.Namespace
        Namespace containing the needed arguments, these are:
        - `sample_regex`: A dictionary of regex strings to search for the posterior files.
        - `preferred_labels`: A list of preferred labels to search for in the posterior files.
        - `parameters`: A list of parameters to extract from the posteriors.
        - `mass_prior`: The mass prior used in initial sampling.
        - `distance_prior`: The distance prior used in initial sampling.
        - `spin_prior`: The spin prior used in initial sampling.
        - `max_redshift`: The maximum redshift allowed in the sample.
    save_meta_data: bool
        Whether to write meta data about the loaded results to plain-text files.
    ignore: list
        List of strings to ignore in the file names to filter unwanted events.

    Returns
    -------
    posteriors: dict
        Dictionary of `pd.DataFrame` posteriors.
    """
    posteriors = dict()
    meta_data = dict()
    logger.info("Loading posteriors...")
    for label, regex in args.sample_regex.items():
        posts, meta = _load_batch_of_meta_files(
            regex=regex,
            label=label,
            labels=args.preferred_labels,
            keys=args.parameters,
            ignore=ignore,
        )
        posts = evaluate_prior(posts, args=args, dataset=label, meta=meta)
        posteriors.update(posts)
        meta_data.update(meta)
    if save_meta_data:
        with open(os.path.join(args.run_dir, "data", "event_data.json"), "w") as ff:
            json.dump(meta_data, ff)
    for key in args.parameters:
        for name in posteriors:
            if key not in posteriors[name]:
                raise KeyError(f"{key} not found for {name}")
    logger.info(f"Loaded {len(posteriors)} posteriors.")
    return posteriors


def plot_summary(posteriors: list, events: list, args):
    """
    Plot a summary of the posteriors for each parameter.

    Parameters
    ----------
    posteriors: list
        List of `pd.DataFrame` posteriors.
    events: list
        Names for each event.
    args
    """
    posteriors = posteriors[::-1]
    events = events[::-1]
    plot_dir = os.path.join(args.run_dir, "data")
    plot_parameters = args.parameters + ["prior"]
    n_cols = len(plot_parameters)
    fig, axes = plt.subplots(
        ncols=n_cols, figsize=(5 * n_cols, len(posteriors)), sharey=True
    )
    for parameter, axis in zip(plot_parameters, axes):
        data = [post[parameter] for post in posteriors]
        plt.sca(axis)
        plt.violinplot(data, vert=False)
        plt.xlabel(parameter.replace("_", " "))
        plt.xlim(np.min(data), np.max(data))
        if parameter == "prior":
            plt.xscale("log")
    plt.ylim(0.5, len(events) + 0.5)
    plt.yticks(np.arange(1, len(events) + 1), events, rotation=90)
    plt.tight_layout()
    plt.savefig(f"{plot_dir}/events.png")
    plt.close(fig)


def gather_posteriors(args, save_meta_data=True):
    """
    Load in posteriors from files according to the command-line arguments.

    Parameters
    ----------
    args: argparse.Namespace
        Command-line arguments
    save_meta_data: bool
        Whether to write meta data about the loaded results to plain-text files.

    Returns
    -------
    posts: list
        List of `pd.DataFrame` posteriors.
    events: list
        Event labels
    """
    posteriors = load_all_events(
        args, save_meta_data=save_meta_data, ignore=args.ignore
    )
    posts = list()
    events = list()
    filenames = list()
    for filename in posteriors.keys():
        event = re.findall(r"(GW\d{6}_\d{6}|S\d{6}[a-z]*|GW\d{6})", filename)[-1]
        if event in events:
            logger.warning(f"Duplicate event {event} found, ignoring {filename}.")
            continue
        posts.append(posteriors[filename])
        events.append(event)
        filenames.append(filename)
    if save_meta_data:
        logger.info(f"Outdir is {args.run_dir}")
        with open(
            f"{args.run_dir}/data/{args.data_label}_posterior_files.txt", "w"
        ) as ff:
            for event, filename in zip(events, filenames):
                ff.write(f"{event}: {filename}\n")
    n_samples = args.samples_per_posterior
    for post in posts:
        n_samples = min(len(post), n_samples)
    logger.info(f"Downsampling to {n_samples} samples per posterior")
    downsampled_posts = [
        post.sample(n_samples, random_state=args.collection_seed) for post in posts
    ]
    return downsampled_posts, events


def resolve_arguments(args):
    """
    - Make sure there are no incompatible arguments.
    - Resolve any deprecated arguments with their corresponding updates if possible.
    - Disable prior terms for parameters that aren't being fit.
    """
    if args.mass_prior.lower() == "flat-detector":
        logger.warning(
            "The 'flat-detector' mass prior specification is deprecated, "
            "use 'flat-detector-components' instead."
        )
        args.mass_prior = "flat-detector-components"
    elif args.mass_prior.lower() == "chirp-mass":
        logger.warning(
            "The 'chirp-mass' mass prior specification is deprecated, "
            "use 'flat-detector-chirp-mass-ratio' instead."
        )
        args.mass_prior = "flat-detector-chirp-mass-ratio"

    mass_parameters = {
        "mass_1",
        "mass_1_detector",
        "mass_2",
        "mass_2_detector",
        "chirp_mass",
        "chirp_mass_detector",
        "mass_ratio",
    }
    fitted_masses = mass_parameters.intersection(args.parameters)
    if len(fitted_masses) > 2:
        logger.warning(
            "More than two mass parameters specified, this may lead to issues with the prior."
        )
    elif len(fitted_masses) == 1:
        logger.warning(
            "Only one mass parameter specified, this may lead to issues with the prior."
        )
    elif len(fitted_masses) == 0:
        args.mass_prior = "None"

    if (
        "redshift" not in args.parameters
        and "luminosity_distance" not in args.parameters
    ):
        args.distance_prior = "None"

    spin_parameters = {
        "a_1",
        "a_2",
        "cos_tilt_1",
        "cos_tilt_2",
        "chi_1",
        "chi_2",
        "chi_eff",
        "chi_p",
    }
    fitted_spins = spin_parameters.intersection(args.parameters)
    if len(fitted_spins) == 0:
        args.spin_prior = "None"
    args.sample_regex = convert_arg_to_dict(args.sample_regex)
    prior_dict = {
        key: vars(args)[key] for key in ["mass_prior", "spin_prior", "distance_prior"]
    }
    for param in prior_dict:
        if "{" not in prior_dict[param]:
            prior_dict[param] = {
                dataset: prior_dict[param] for dataset in args.sample_regex
            }
        else:
            prior_dict[param] = convert_arg_to_dict(prior_dict[param])
    vars(args).update(prior_dict)


def convert_arg_to_dict(arg):
    """
    Convert a string argument to a dictionary. Not in-place.
    gwpopulation_pipe strips quotes from the regex string, so we need to add them back in,
    this assumes that there are no internal braces and spaces after all ':' and ','
    delimiting entries.

    Parameters
    ----------
    arg: str
        Arg that should be converted to a dictionary.

    Returns
    -------
    arg_dict: dict
        Dictionary representation of the input string.
    """
    try:
        regex_str = arg
        if '"' not in regex_str:

            regex_str = (
                regex_str.replace("{", '{"')
                .replace(":", '":"')
                .replace(", ", '", "')
                .replace("}", '"}')
                .replace(" ", "")
            )
        arg_dict = json.loads(regex_str)
    except json.decoder.JSONDecodeError:
        arg_dict = convert_string_to_dict(arg)
    return arg_dict


def main():
    parser = create_parser()
    args = parser.parse_args()
    resolve_arguments(args)

    set_backend("numpy")

    os.makedirs(f"{args.run_dir}/data", exist_ok=True)
    if args.injection_file is not None or args.sample_from_prior:
        posts = simulate_posteriors(args=args)
        events = [str(ii) for ii in range(len(posts))]
    else:
        posts, events = gather_posteriors(args=args)
    logger.info(f"Using {len(posts)} events, final event list is: {', '.join(events)}.")
    posterior_file = f"{args.data_label}.pkl"
    logger.info(f"Saving posteriors to {posterior_file}")
    filename = os.path.join(args.run_dir, "data", posterior_file)
    pd.to_pickle(posts, filename)
    if args.plot:
        plot_summary(posts, events, args)
    if args.vt_file is not None:
        dump_injection_data(args)



# Adapted from transformations derived in Iwaya et al. 2024

import numpy as np
from gwpopulation.utils import to_numpy, xp

try:
    import jax
except ImportError:
    pass
from scipy.special import spence as scipyspence
from .utils import maybe_jit

# Relative tolerance for the series
TOL = 2.220446092504131e-16
PISQ_6 = 1.6449340668482264365
MAX_ITER = 500


def I1(chieff, chip, q):
    x1max = xp.minimum(
        xp.sqrt(q**2 - ((4 + 3 * q) / (3 + 4 * q)) ** 2 * chip**2),
        (1 + q) * chieff + xp.sqrt(1 - chip**2),
    )
    x1min = xp.maximum(
        -xp.sqrt(q**2 - ((4 + 3 * q) / (3 + 4 * q)) ** 2 * chip**2),
        (1 + q) * chieff - xp.sqrt(1 - chip**2),
    )
    cond1 = chip > 0
    cond2 = q >= (4 + 3 * q) * chip / (3 + 4 * q)
    cond3 = x1max >= x1min

    Fterm = F(x1max, (1 + q) * chieff, chip, (4 + 3 * q) / (3 + 4 * q) * chip, q) - F(
        x1min, (1 + q) * chieff, chip, (4 + 3 * q) / (3 + 4 * q) * chip, q
    )
    return xp.where(cond1 & cond2 & cond3, (1 + q) / (8 * q) * Fterm, 0)


def I2(chieff, chip, q):
    x2max = xp.minimum(q, (1 + q) * chieff + xp.sqrt(1 - chip**2))
    x2min = xp.maximum(-q, (1 + q) * chieff - xp.sqrt(1 - chip**2))

    cond1 = chip > 0
    cond2 = chip < 1
    cond3 = x2max >= x2min

    Fterm = F(x2max, (1 + q) * chieff, chip, 0, q) - F(
        x2min, (1 + q) * chieff, chip, 0, q
    )
    return xp.where(cond1 & cond2 & cond3, -(1 + q) / (8 * q) * Fterm, 0)


def I3(chieff, chip, q):
    x3max = xp.minimum(
        xp.sqrt(1 - chip**2),
        (1 + q) * chieff
        + xp.sqrt(q**2 - ((4 + 3 * q) / (3 + 4 * q)) ** 2 * chip**2),
    )
    x3min = xp.maximum(
        -xp.sqrt(1 - chip**2),
        (1 + q) * chieff
        - xp.sqrt(q**2 - ((4 + 3 * q) / (3 + 4 * q)) ** 2 * chip**2),
    )

    cond1 = chip > 0
    cond2 = q >= (4 + 3 * q) * chip / (3 + 4 * q)
    cond3 = x3max > x3min

    Fterm = F(x3max, (1 + q) * chieff, (4 + 3 * q) / (3 + 4 * q) * chip, chip, 1) - F(
        x3min, (1 + q) * chieff, (4 + 3 * q) / (3 + 4 * q) * chip, chip, 1
    )

    return xp.where(
        cond1 & cond2 & cond3, (1 + q) / (8 * q) * (4 + 3 * q) / (3 + 4 * q) * Fterm, 0
    )


def I4(chieff, chip, q):
    x4max = xp.minimum(
        1,
        (1 + q) * chieff
        + xp.sqrt(q**2 - ((4 + 3 * q) / (3 + 4 * q)) ** 2 * chip**2),
    )
    x4min = xp.maximum(
        -1,
        (1 + q) * chieff
        - xp.sqrt(q**2 - ((4 + 3 * q) / (3 + 4 * q)) ** 2 * chip**2),
    )

    cond1 = chip > 0
    cond2 = q >= (4 + 3 * q) * chip / (3 + 4 * q)
    cond3 = x4max > x4min

    Fterm = F(x4max, (1 + q) * chieff, (4 + 3 * q) / (3 + 4 * q) * chip, 0, 1) - F(
        x4min, (1 + q) * chieff, (4 + 3 * q) / (3 + 4 * q) * chip, 0, 1
    )

    return xp.where(
        cond1 & cond2 & cond3, -(1 + q) / (8 * q) * (4 + 3 * q) / (3 + 4 * q) * Fterm, 0
    )


def F(x, a, b, c, d):
    return G(x / b, a / b, c / b) + xp.log(b**2 / d**2) * (
        xp.arctan((x - a) / b) + xp.arctan(a / b)
    )


def G(x, alpha, beta):
    pre = xp.where(x >= 0, 1, -1)
    alpha = xp.where(x >= 0, alpha, -alpha)
    x = xp.where(x >= 0, x, -x)
    g1 = g(x, alpha, beta)
    g2 = g(x, alpha, -beta)
    g3 = -g(0, alpha, beta)
    g4 = -g(0, alpha, -beta)

    out = pre * xp.imag(g1 + g2 + g3 + g4)
    return out


def g(x, alpha, beta):
    cond1 = xp.abs(beta) < 1
    cond2 = (beta == 1) & (alpha <= 0)
    x_ = xp.where((x == 0) & (beta == 0), 0.01, x)
    ret = xp.nan_to_num(
        xp.where(
            cond1,
            xp.log(x_ - beta * 1j)
            * xp.log((alpha - x_ + 1j) / (alpha + 1j - beta * 1j))
            + Li2((x_ - beta * 1j) / (alpha + 1j - beta * 1j)),
            xp.where(
                cond2,
                0.5 * (xp.log(x_ - alpha - 1j)) ** 2 + Li2(-alpha / (x_ - alpha - 1j)),
                xp.log(alpha + 1j - beta * 1j) * xp.log(alpha - x_ + 1j)
                - Li2((alpha - x_ + 1j) / (alpha + 1j - beta * 1j)),
            ),
        )
    )
    return xp.where((x == 0) & (beta == 0), 0, ret)


def jaxspence(z):
    """
    From scipy.special.spence's implementation:
    Compute Spence's function for complex arguments. The strategy is:
    - If z is close to 0, use a series centered at 0.
    - If z is far away from 1, use the reflection formula

    spence(z) = -spence(z/(z - 1)) - pi**2/6 - ln(z - 1)**2/2

    to move close to 1.
    - If z is close to 1, use a series centered at 1.

    """
    return jax.lax.cond(
        xp.abs(z) < 0.5,
        lambda z: cspence_series0(z),
        lambda z: jax.lax.cond(
            xp.abs(1 - z) > 1,
            lambda z: -cspence_series1(z / (z - 1)) - PISQ_6 - 0.5 * xp.log(z - 1) ** 2,
            lambda z: cspence_series1(z),
            z,
        ),
        z,
    )


def cspence_series0(z):
    """
    A series centered at z = 0; see

    http://functions.wolfram.com/10.07.06.0005.02

    """
    z_ = xp.where(z == 0, 0.01, z)

    def condition(args):
        n, zfac, term1, sum1, term2, sum2 = args
        return (
            (xp.abs(term1) > TOL * xp.abs(sum1)) | (xp.abs(term2) > TOL * xp.abs(sum2))
        ) & (n < MAX_ITER)

    def body(args):
        n, zfac, term1, sum1, term2, sum2 = args
        zfac *= z_
        term1 = zfac / n**2
        sum1 += term1
        term2 = zfac / n
        sum2 += term2
        return (n + 1, zfac, term1, sum1, term2, sum2)

    def body_fori(i, args):
        return body(args)

    n, zfac, term1, sum1, term2, sum2 = jax.lax.fori_loop(
        1, MAX_ITER, body_fori, (1, 1, 0, 0, 0, 0)
    )

    # n, zfac, term1, sum1, term2, sum2 = jax.lax.while_loop(
    # condition, body, (1, 1, xp.inf, 0, xp.inf, 0)
    # )
    return xp.where(z == 0, PISQ_6, PISQ_6 - sum1 + xp.log(z_) * sum2)


def cspence_series1(z):
    """
    A series centered at z = 1 which enjoys faster convergence than
    the Taylor series. The number of terms used comes from
    bounding the absolute tolerance at the edge of the radius of
    convergence where the sum is O(1).

    """

    z_ = xp.where(z == 1, 1 - 1e-10, z)

    z_ = 1 - z_
    zz = z_**2

    def condition(args):
        n, zfac, res, term = args
        return (xp.abs(term) > TOL * xp.abs(res)) & (n < MAX_ITER)

    def body(args):
        n, zfac, res, _ = args
        zfac *= z_
        term = ((zfac / n**2) / (n + 1) ** 2) / (n + 2) ** 2
        res += term
        return (n + 1, zfac, res, term)

    def body_fori(i, args):
        return body(args)

    n, zfac, res, term = jax.lax.fori_loop(1, MAX_ITER, body_fori, (1, 1, 0, xp.inf))
    # n, zfac, res, term = jax.lax.while_loop(condition, body, (1, 1, 0, xp.inf))
    res *= 4 * zz
    res += 4 * z_ + 5.75 * zz + 3 * (1 - zz) * xp.log(1 - z_)
    res /= 1 + 4 * z_ + zz
    return xp.where(z == 1, 0.0, res)


def Li2(z):
    if "jax" in xp.__name__:
        spence = jax.vmap(jaxspence)
    else:
        from scipy.special import spence
    z = xp.atleast_1d(z)
    return spence(1 - z)


@maybe_jit
def prior_chieff_chip_isotropic(chieff, chip, q, amax=1):
    chieff = chieff / amax
    chip = chip / amax
    return (
        I1(chieff, chip, q)
        + I2(chieff, chip, q)
        + I3(chieff, chip, q)
        + I4(chieff, chip, q)
    ) / amax**2



import os

from gwpopulation.conversions import convert_to_beta_parameters
from gwpopulation.hyperpe import HyperparameterLikelihood, xp
from gwpopulation.models.spin import (
    iid_spin,
    iid_spin_magnitude_beta,
    independent_spin_magnitude_beta,
)


def get_cosmology(name):
    from wcosmo.astropy import FlatLambdaCDM, available

    if name.lower() == "planck15_lal":
        return FlatLambdaCDM(H0=67.90, Om0=0.3065, name="Planck15_LAL")
    elif name in available.keys():
        return available[name]
    else:
        raise ValueError(f"Unknown cosmology {name}")


def prior_conversion(parameters):
    """Wrapper around conversion for prior constraints"""
    for key in ["amax", "amax_1", "amax_2"]:
        if key not in parameters:
            parameters[key] = 1
    parameters, _ = convert_to_beta_parameters(parameters)
    return parameters


def get_path_or_local(path: str) -> str:
    """
    Check if a path exists, if not check if the basename exists, if not raise an error.

    Parameters
    ----------
    path: str
        The path to check.

    Returns
    -------
    path: str
        The path to use.

    Raises
    ------
    ValueError
        If neither the path nor the basename exist.
    """
    if os.path.exists(path):
        return path
    elif os.path.exists(os.path.basename(path)):
        return os.path.basename(path)
    else:
        raise ValueError(f"Cannot find {path} or {os.path.basename(path)}")


def chi_eff_from_components(a_1, cos_tilt_1, a_2, cos_tilt_2, mass_ratio):
    """
    Compute the effective spin parameter from the component spins.

    Parameters
    ----------
    a_1: array-like
        The magnitude of the primary spin.
    cos_tilt_1: array-like
        The cosine of the tilt of the primary spin.
    a_2: array-like
        The magnitude of the secondary spin.
    cos_tilt_2: array-like
        The cosine of the tilt of the secondary spin.
    mass_ratio: array-like
        The mass ratio of the binary.

    Returns
    -------
    chi_eff: array-like
        The effective spin parameter.
    """
    return (a_1 * cos_tilt_1 + mass_ratio * a_2 * cos_tilt_2) / (1 + mass_ratio)


def chi_p_from_components(a_1, cos_tilt_1, a_2, cos_tilt_2, mass_ratio):
    """
    Compute the precession spin parameter from the component spins.

    Parameters
    ----------
    a_1: array-like
        The magnitude of the primary spin.
    cos_tilt_1: array-like
        The cosine of the tilt of the primary spin.
    a_2: array-like
        The magnitude of the secondary spin.
    cos_tilt_2: array-like
        The cosine of the tilt of the secondary spin.
    mass_ratio: array-like
        The mass ratio of the binary.

    Returns
    -------
    chi_p: array-like
        The precession spin parameter.
    """
    sin_tilt_1 = xp.sqrt(1 - cos_tilt_1**2)
    sin_tilt_2 = xp.sqrt(1 - cos_tilt_2**2)
    return xp.maximum(
        a_1 * sin_tilt_1,
        (3 + 4 * mass_ratio) / (4 + 3 * mass_ratio) * mass_ratio * a_2 * sin_tilt_2,
    )


KNOWN_ARGUMENTS = {
    iid_spin: ["mu_chi", "sigma_chi", "xi_spin", "sigma_spin"],
    iid_spin_magnitude_beta: ["mu_chi", "sigma_chi"],
    independent_spin_magnitude_beta: [
        "mu_chi_1",
        "mu_chi_2",
        "sigma_chi_1",
        "sigma_chi_2",
    ],
}


class MinimumEffectiveSamplesLikelihood(HyperparameterLikelihood):
    def _compute_per_event_ln_bayes_factors(self, return_uncertainty=True):
        """
        Compute the per event ln Bayes factors and associated variance.

        This method imposes a condition that the number of effective
        samples per Monte Carlo integral must be at least as much
        as the total number of events. Otherwise the lnBF is set to
        - infinity.

        Returns
        -------
        ln_per_event_bfs: array-like
            The ln BF per event subject to having a sufficient number
            of independent samples.
        variance: array-like
            The variances (uncertainties) in the ln BF per event. Only
            returned if `return_uncertainty` is True. This output will
            generally not be used for convergence criteria, as this function
            already enforces a threshold on effective number of samples.
        """
        (
            per_event_bfs,
            n_effectives,
            variance,
        ) = self.per_event_bayes_factors_and_n_effective_and_variances()
        per_event_bfs *= n_effectives > self.n_posteriors
        if return_uncertainty:
            return xp.log(per_event_bfs), variance
        else:
            return xp.log(per_event_bfs)

    def per_event_bayes_factors_and_n_effective_and_variances(self):
        """
        Called by `_compute_per_event_ln_bayes_factors` to compute the
        per event BFs, effective number of samples for each event's computed
        BF, and the associated uncertainty (variance) in the *ln* BF. Computes
        same qunatities as superclass function `_compute_per_event_ln_bayes_factors`
        but additionally provides the effective sample size.

        Returns
        -------
        per_event_bfs: array-like
            The BF per event, computed by reweighting single-event likelihood
            samples into the `hyper_prior` model.
        n_effectives: array-like
            The effective sample size for each Monte Carlo sum computation of the BFs.
            The BF is computed for each event, so this array has length n_events.
        variance: array-like
            The variances (uncertainties) in the ln BF per event.
        """
        weights = self.hyper_prior.prob(self.data) / self.sampling_prior
        per_event_bfs = xp.sum(weights, axis=-1)
        n_effectives = xp.nan_to_num(per_event_bfs**2 / xp.sum(weights**2, axis=-1))
        per_event_bfs /= self.samples_per_posterior
        square_expectation = xp.mean(weights**2, axis=-1)
        variance = (square_expectation - per_event_bfs**2) / (
            self.samples_per_posterior * per_event_bfs**2
        )
        return per_event_bfs, n_effectives, variance

    def per_event_bayes_factors_and_n_effective(self):
        (
            per_event_bfs,
            n_effectives,
            _,
        ) = self.per_event_bayes_factors_and_n_effective_and_variances()
        return per_event_bfs, n_effectives


def maybe_jit(func):
    from gwpopulation.backend import __backend__

    if __backend__ == "jax":
        from jax import jit

        return jit(func)
    else:
        return func



"""
Translation of the cython implementation of complex Spence
from scipy to work with jax.
Adapted from https://github.com/scipy/scipy/blob/8971d5e9b72931987b7d3c5a25da1a8e7e5485d0/scipy/special/_spence.pxd
"""

import jax.numpy as jnp

TOL = 1e-15

def cspence_series0(z):
    """
    Small z branch, uses a series expansion about :math:`z = 0`
    for :math:`|z| < 0.5`.
    """

    def body_func(val):
        z, zfac, n, sum1, sum2, term1, term2 = val
        zfac *= z
        term1 = zfac / n**2
        sum1 += term1
        term2 = zfac / n
        sum2 += term2
        n = n + 1
        return jnp.array([z, zfac, n, sum1, sum2, term1, term2])
    
    def cond_fun(val):
        z, zfac, n, sum1, sum2, term1, term2 = val
        return (n > 500) | (
            (abs(term1) > TOL * abs(sum1))
            & (abs(term2) > TOL * abs(sum2))
        )

    val = jnp.array([z, 1.0, 1, 0, 0, 1, 1])
    result = while_loop(cond_fun, body_func, val)
    sum1 = result[3]
    sum2 = result[4]
    return np.pi**2 / 6 - sum1 + jnp.log(z) * sum2


def cspence_series1(z):
    """
    Middle branch, an expansion around :math:`z = 1`
    for :math:`|z| > 0.5` and :math:`|1 - z| > 1`.
    """

    def body_func(val):
        z, zfac, n, res, term = val
        zfac *= z
        term = ((zfac/n**2)/(n + 1)**2)/(n + 2)**2
        res += term
        n = n + 1
        return jnp.array([z, zfac, n, res, term])
    
    def cond_fun(val):
        z, zfac, n, res, term = val
        return (n > 500) | (abs(term) > 1e-15 * abs(res))

    z = 1 - z
    zz = z**2
    val = jnp.array([z, 1.0, 1, 0, 1])
    result = while_loop(cond_fun, body_func, val)
    res = result[3] * 4 * zz
    res += 4 * z + 5.75 * zz + 3 * (1 - zz) * jnp.log(1 - z)
    res /= 1 + 4 * z + zz
    return res


def cspence_series2(z):
    """
    Large :math:`z` branch for :math:`|z| > 0.5` and
    :math:`|1 - z| > 1`. Uses a reflection expression.
    """
    return (
        -cspence_series1(z / (z - 1))
        - np.pi**2 / 6
        - jnp.log(z - 1)**2 / 2
    )


def cspence(z):
    """
    Return the Spence dilogarithm for complex input using
    branches dependent on the value of the input.
    """
    return cond(
        abs(z) < 0.5,
        cspence_series0,
        lambda z: cond(
            abs(1 - z) > 1,
            cspence_series2,
            cspence_series1,
            z,
        ),
        z,
    )



"""
Functions for generating simulated event posteriors.
The simulated posteriors will not accurately represent real observational uncertainties.

The module provides the `gwpopulation_pipe_simulate_posteriors` executable.

In order to use the primary function you will need a class that provides
various attributes specified in the `gwpopulation_pipe` parser.
"""

import os
import sys

import numpy as np
import pandas as pd
from bilby.core.prior import TruncatedGaussian
from bilby.core.utils import logger
from bilby.gw.cosmology import get_cosmology
from bilby.gw.prior import UniformSourceFrame
from gwpopulation.backend import set_backend
from gwpopulation.conversions import convert_to_beta_parameters
from gwpopulation.utils import to_numpy, xp

from .data_analysis import load_model, load_vt
from .main import MASS_MODELS
from .main import create_parser as create_main_parser

BOUNDS = dict(
    mass_1=(2, 100),
    mass_ratio=(0, 1),
    a_1=(0, 1),
    a_2=(0, 1),
    cos_tilt_1=(-1, 1),
    cos_tilt_2=(-1, 1),
    redshift=(0, 2.3),
)

cosmology = get_cosmology()

PRIOR_VOLUME = (
    (BOUNDS["mass_1"][1] - BOUNDS["mass_1"][0]) ** 2
    * (BOUNDS["a_1"][1] - BOUNDS["a_1"][0])
    * (BOUNDS["a_2"][1] - BOUNDS["a_2"][0])
    * (BOUNDS["cos_tilt_1"][1] - BOUNDS["cos_tilt_1"][0])
)


def create_parser():
    parser = create_main_parser()
    parser.add_argument(
        "--models",
        action="append",
        help="Model function to evaluate, default is "
        "two component mass and iid spins.",
    )
    parser.add_argument(
        "--vt-model",
        type=str,
        default="",
        help="Function to generate VT object, should be in "
        "vt_helper.py, default is mass only.",
    )
    return parser


def draw_true_values(model, vt_model=None, n_samples=40):
    """
    Draw samples from the probability density provided by the input model.

    Parameters
    ----------
    model: bilby.hyper.model.Model
        Population model to evaluate
    vt_model: gwpopulation.vt.GridVT, optional
        Model to compute selection function
    n_samples: int
        Number of samples to draw

    Returns
    -------
    total_samples: `pd.DataFrame`
        Data containing the true values.

    """
    if vt_model is None:
        vt_model = lambda x: 1
    else:
        raise NotImplementedError("Sensitive volume not implemented.")
    total_samples = pd.DataFrame()
    n_per_iteration = n_samples * 10000
    while True:
        data = _draw_from_prior(n_samples=n_per_iteration)
        prob = model.prob(data)
        prob *= vt_model(data)
        data = pd.DataFrame({key: to_numpy(data[key]) for key in data})
        data["prob"] = to_numpy(prob)
        total_samples = total_samples.append(data)
        max_prob = np.max(total_samples["prob"])
        total_samples = total_samples[total_samples["prob"] > 0]
        total_samples = total_samples[
            total_samples["prob"] > max_prob / n_per_iteration
        ]
        keep = np.array(total_samples["prob"]) >= np.random.uniform(
            0, max_prob, len(total_samples)
        )
        if sum(keep) > n_samples:
            total_samples = total_samples.iloc[keep]
            break
        logger.info(f"Drawing events is very inefficient.")
    total_samples = total_samples[:n_samples]
    del total_samples["prob"]
    return total_samples


def simulate_posterior(sample, fractional_sigma=0.1, n_samples=10000):
    """
    Simulate a posterior distribution given an input sample.

    This assumes uncorrelated uncertainties between different parameters.

    Parameters
    ----------
    sample: dict
        The true parameter values of the event to simulate a posterior for.
    fractional_sigma: float
        Fractional uncertainty on each parameter.
    n_samples: int
        The number of samples to draw.

    Returns
    -------
    posterior: `pd.DataFrame`
        The simulated posterior samples.

    """
    posterior = pd.DataFrame()
    for key in sample:
        if key in BOUNDS:
            bound = BOUNDS[key]
        else:
            bound = (-np.inf, np.inf)
        sigma = sample[key] * fractional_sigma
        new_true = TruncatedGaussian(
            mu=sample[key], sigma=sigma, minimum=bound[0], maximum=bound[1]
        ).sample()
        posterior[key] = TruncatedGaussian(
            mu=new_true, sigma=sigma, minimum=bound[0], maximum=bound[1]
        ).sample(n_samples)
    posterior = posterior[
        (BOUNDS["mass_1"][0] / posterior["mass_1"] <= posterior["mass_ratio"])
        & (posterior["mass_ratio"] <= BOUNDS["mass_ratio"][1])
    ]

    posterior["prior"] = 1 / PRIOR_VOLUME
    return posterior


def _draw_from_prior(n_samples):
    data = dict()
    for key in BOUNDS:
        data[key] = xp.random.uniform(BOUNDS[key][0], BOUNDS[key][1], n_samples)
    data["redshift"] = xp.asarray(
        UniformSourceFrame(
            minimum=BOUNDS["redshift"][0],
            maximum=BOUNDS["redshift"][1],
            name="redshift",
        ).sample(n_samples)
    )
    return data


def simulate_posterior_from_prior(n_samples=10000):
    """
    Draw samples from the posterior distribution assuming the data are
    non-informative, e.g., sample from the prior distribution.

    Parameters
    ----------
    n_samples: int
        The number of samples to simulate.

    Returns
    -------
    posterior: `pd.DataFrame`
        The simulated samples.

    """
    data = _draw_from_prior(n_samples=n_samples)
    posterior = pd.DataFrame({key: to_numpy(data[key]) for key in data})
    posterior["prior"] = (
        UniformSourceFrame(
            minimum=BOUNDS["redshift"][0],
            maximum=BOUNDS["redshift"][1],
            name="redshift",
        ).prob(posterior["redshift"])
        / PRIOR_VOLUME
    )
    return posterior


def simulate_posteriors(args):
    """
    Simulate a set of posterior samples gives the command-line arguments.

    Parameters
    ----------
    args:
        The command-line arguments.

    Returns
    -------
    posteriors: list
        The simulated posterior samples.
    """
    posteriors = list()
    if args.sample_from_prior:
        logger.info("Drawing prior samples for all simulated events.")
        for ii in range(args.n_simulations):
            posteriors.append(
                simulate_posterior_from_prior(n_samples=args.samples_per_posterior)
            )
    else:
        injection_parameters = dict(
            pd.read_json(args.injection_file).iloc[args.injection_index]
        )
        injection_parameters, _ = convert_to_beta_parameters(
            parameters=injection_parameters
        )
        args.models = list()
        args.models.append(MASS_MODELS[args.mass_models[0]])
        args.models.append(f"{args.magnitude_models[0]}_spin_magnitude")
        args.models.append(f"{args.tilt_models[0]}_spin_orientation")
        model = load_model(args=args)
        model.parameters.update(injection_parameters)
        logger.info(model.parameters)
        if not getattr(args, "vt_model", "") == "":
            vt_model = load_vt(args=args)
        else:
            vt_model = None
        true_values = draw_true_values(
            model=model, vt_model=vt_model, n_samples=args.n_simulations
        )
        logger.info("Simulating posteriors for all events.")
        for ii in range(args.n_simulations):
            posteriors.append(
                simulate_posterior(
                    dict(true_values.iloc[ii]), n_samples=args.samples_per_posterior
                )
            )
    if not os.path.exists(os.path.join(args.run_dir)):
        os.mkdir(os.path.join(args.run_dir))
    if not os.path.exists(os.path.join(args.run_dir, "data")):
        os.mkdir(os.path.join(args.run_dir, "data"))
    with open(
        os.path.join(args.run_dir, "data", f"{args.data_label}_posterior_files.txt"),
        "w",
    ) as ff:
        for ii in range(args.n_simulations):
            ff.write(f"{ii}:{ii}")
    return posteriors


def main():
    parser = create_parser()
    args, _ = parser.parse_known_args(sys.argv[1:])

    set_backend(args.backend)

    simulate_posteriors(args)
